#NautilusPathBar {
  box-shadow: inset 0 0 0 1px var(--border-color);
}

#NautilusPathButton:not(:hover):not(:drop(active)),
#NautilusPathButton.current-dir:not(:drop(active)) {
  box-shadow: none;
}

.disk-space-free {
  color: color-mix(in srgb, currentColor 30%, transparent);
}

.nautilus-list-view columnview > listview > row.activatable:hover,
.nautilus-grid-view gridview > child.activatable:hover {
  box-shadow: inset 0 0 0 1px var(--border-color);
}

.nautilus-list-view columnview > listview > row:selected,
.nautilus-grid-view gridview > child:selected {
  box-shadow: inset 0 0 0 1px color-mix(in srgb, var(--accent-color) 60%, transparent);
}

